import sys
from typing import Optional, List

from envvarconf.inspection import get_settings_variables
from envvarconf.printing import get_tty_width, truncatestring
from envvarconf.validation import validate_definition, validation_errors

from envvarconf.loaders.base import BaseLoader


class BaseSettings:
    def __init__(self, appname: str = ""):
        self.appname = appname
        validate_definition(self)

    def load(self, loaders: List[BaseLoader], failfast=True, truncate_width: Optional[int] = None):
        """
        Load settings using `loader`

        :param loaders: List of loaders instance, see `loaders` module
        :param truncate_width: Truncate output width, default using terminal width
        :param failfast: If settings is not valid print help then exit, else do nothing
        """

        for loader in loaders:
            loader.load(self)

        errors = validation_errors(self)
        if errors:
            print("There is errors in settings")
            print("\n".join([f" * {x}" for x in errors]))
            print()

            if failfast:
                self.print_detailed_settings(truncate_width=truncate_width)
                sys.exit(1)

    def print_detailed_settings(self, truncate_width: Optional[int] = None):
        """
        Pring each settings variable and value

        :param truncate_width: Truncate output width, default using terminal width
        """
        if truncate_width is None:
            truncate_width = get_tty_width()
        print(f"{self.appname or 'Application'} settings:")

        for varname, vartype in get_settings_variables(self).items():
            varval = str(getattr(self, varname, "NOT DEFINED!"))
            print(truncatestring(f"{varname}:{vartype.__name__} = {varval}", width=truncate_width, placeholder="..."))

    def __str__(self):
        vs = [f"{var_name}={getattr(self, var_name, '?')}" for var_name in get_settings_variables(self).keys()]
        return f"<{self.__class__.__name__} {', '.join(vs)}>"



from typing import Type, Any, TYPE_CHECKING

from envvarconf.converters.base import ConverterBase, ALLOWED_TYPES


if TYPE_CHECKING:
    from envvarconf import BaseSettings


class BaseLoader:
    """
    Abstract base loader class
    """
    converter: ConverterBase = ConverterBase()

    @classmethod
    def set_var(cls, settings: 'BaseSettings', varname: str, raw_value: Any, vartype: Type):
        assert vartype in ALLOWED_TYPES, f"Variable {varname} marked {vartype}, that not allowed for settings use"
        setattr(
            settings,
            varname,
            cls.converter.convert(raw_value, vartype)
        )

    def load(self, settings: 'BaseSettings') -> 'BaseSettings':
        raise NotImplementedError



from typing import List

from envvarconf.converters.base import ALLOWED_TYPES
from envvarconf.inspection import get_settings_variables


def validate_definition(settings):
    """
    Check class definition is valid

    :raises: TypeError when class definition is wrong
    """
    for varname, vartype in get_settings_variables(settings).items():
        if vartype not in ALLOWED_TYPES:
            raise TypeError(f"Variable {varname} marked {vartype}, that not allowed for settings use")


def validation_errors(settings) -> List[str]:
    """
    Get settings validation_errors

    :return: List of strings, representing errors detail
    """
    errs = []
    for varname, vartype in get_settings_variables(settings).items():
        if not hasattr(settings, varname):
            errs.append(f"{varname} is not defined")
    return errs



import os

from envvarconf import BaseSettings
from envvarconf.converters.string_converter import StringConverter
from envvarconf.inspection import get_settings_variables
from envvarconf.loaders.base import BaseLoader


class Loader(BaseLoader):
    converter = StringConverter()

    def load(self, settings: BaseSettings):
        for varname, vartype in get_settings_variables(settings).items():
            if varname in os.environ:
                self.set_var(settings, varname, raw_value=os.environ[varname], vartype=vartype)



from decimal import Decimal

from envvarconf.converters.string_converter import StringConverter


def test_string_converter():
    converter = StringConverter()

    assert converter.convert("0.001", Decimal) == Decimal("0.001")
    assert converter.convert("hello", str) == "hello"
    assert converter.convert("0.1", float) == 0.1
    assert converter.convert("5", int) == 5



import shutil


def truncatestring(string: str, width=30, placeholder="...") -> str:
    # https://stackoverflow.com/questions/2872512/python-truncate-a-long-string/39017530
    width -= len(placeholder)
    assert width > 0
    return (string[:width] + placeholder) if len(string) > width else string


def get_tty_width(default=80) -> int:
    return shutil.get_terminal_size((default, 20)).columns



from envvarconf import BaseSettings, validation_errors
from envvarconf.loaders import environ


class Settings(BaseSettings):
    SENTRY_DSN: str
    LOGGING_LEVEL: str = 'debug'

    HOST: str = "aaaakehgeiuhgiweurhiuerhf"*200
    PORT: int


def test_settings_invalid():
    settings = Settings(appname="evc_tests")
    settings.load([environ.Loader()], failfast=False)
    errors = validation_errors(settings)
    assert len(errors) == 2
    assert errors == ['SENTRY_DSN is not defined', 'PORT is not defined']


def test_settings_valid(monkeypatch):
    monkeypatch.setenv("SENTRY_DSN", "123")
    monkeypatch.setenv("PORT", "123")

    settings = Settings(appname="evc_tests")
    settings.load([environ.Loader()])
    errors = validation_errors(settings)
    assert len(errors) == 0
    assert settings.SENTRY_DSN == "123"
    assert settings.PORT == 123



from envvarconf.converters.base import ConverterBase


class StringConverter(ConverterBase):
    """
    Simple convert string representation to value
    """

    def convert(self, value: str, to_type):
        return to_type(value)



import logging
import sentry_sdk


logger = logging.getLogger(__name__)


class SentryMixin:
    SENTRY_DSN: str = 'TESTING'

    def initialize_sentry(self, integrations=(), **kwargs):
        """
        Initialize sentry sdk
        """
        if self.SENTRY_DSN == "TESTING":
            logger.debug("Sentry testing mode")
        else:
            logger.debug("Using sentry.")
            sentry_sdk.init(
                dsn=self.SENTRY_DSN,
                integrations=integrations,
                **kwargs
            )



from typing import Type, Dict, get_type_hints, TYPE_CHECKING

if TYPE_CHECKING:
    from envvarconf import BaseSettings


def get_settings_variables(settings: 'BaseSettings') -> Dict[str, Type]:
    annotations = {}
    for inherited_class in reversed(settings.__class__.__mro__):
        if inherited_class != object:
            annotations.update(get_type_hints(inherited_class))
    return annotations



from setuptools import setup, find_packages  # type: ignore
import itertools

VERSION = '1.3'


EXTRAS = {
    "sentry":  ["sentry_sdk"],
}
EXTRAS['all'] = list(itertools.chain(*EXTRAS.values()))

setup(
    name='envvarconf',
    version=VERSION,
    packages=find_packages(),
    url='https://github.com/a1fred/envvarconf',
    license='MIT',
    author='a1fred',
    author_email='demalf@gmail.com',
    description='Safe app configuration from environment variables without extra dependencies',
    classifiers=[
        'Environment :: Console',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    extras_require=EXTRAS,
    test_suite="tests",
)



from decimal import Decimal

ALLOWED_TYPES = [
    str,
    int,
    float,
    Decimal,
]


class ConverterBase:
    def convert(self, value, to_type):
        """
        Convert variable raw value to typed variable

        :param value: variable raw value
        :param to_type: Target type
        :return: typed variable
        """
        raise NotImplementedError



import sys
import logging


class LoggingMixin:
    # CRITICAL 50
    # ERROR 40
    # WARNING 30
    # INFO 20
    # DEBUG 10
    # NOTSET 0
    LOGGING_LEVEL: int = 10

    LOGGING_FORMAT: str = '[%(asctime)s] {%(name)s:%(lineno)d} %(levelname)s - %(message)s'

    def initialize_logging(self):
        """
        Initialize basic logging to stdout
        """
        logging.basicConfig(
            stream=sys.stdout,
            level=self.LOGGING_LEVEL,
            format=self.LOGGING_FORMAT
        )
